
/******************************************************************************
 *
 * � Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/

#import "Swizzle.h"
#import <objc/objc-class.h>

@implementation NSObject (Swizzle)

+ (BOOL)swizzleMethod: (SEL)origSel withMethod: (SEL)altSel
{
#if OBJC_API_VERSION >= 2
    Method origMethod = class_getInstanceMethod(self, origSel);
    
    if (!origMethod)
    {
        NSLog(@"Original method %@ not found for class %@", NSStringFromSelector(origSel), [self className]);
        
        return NO;
    }
    
    Method altMethod = class_getInstanceMethod(self, altSel);
    
    if (!altMethod)
    {
        NSLog(@"Alternate method %@ not found for class %@", NSStringFromSelector(altSel), [self className]);
        
        return NO;
    }
    
    class_addMethod(self, origSel, class_getMethodImplementation(self, origSel), method_getTypeEncoding(origMethod));
    class_addMethod(self, altSel, class_getMethodImplementation(self, altSel), method_getTypeEncoding(altMethod));
    method_exchangeImplementations(class_getInstanceMethod(self, origSel), class_getInstanceMethod(self, altSel));
    
    return YES;
#else
    Method directOriginalMethod = NULL, directAlternateMethod = NULL;
    
    void *iterator = NULL;
    struct objc_method_list *mlist = class_nextMethodList(self, &iterator);
    
    while (mlist)
    {
        int method_index = 0;
        
        for (; method_index < mlist->method_count; ++method_index)
        {
            if (mlist->method_list[method_index].method_name == origSel)
            {
                assert(!directOriginalMethod);
                directOriginalMethod = &mlist->method_list[method_index];
            }
            
            if (mlist->method_list[method_index].method_name == altSel)
            {
                assert(!directAlternateMethod);
                directAlternateMethod = &mlist->method_list[method_index];
            }
        }
        
        mlist = class_nextMethodList(self, &iterator);
    }
    
    if (!directOriginalMethod || !directAlternateMethod)
    {
        Method inheritedOriginalMethod = NULL, inheritedAlternateMethod = NULL;
        
        if (!directOriginalMethod)
        {
            inheritedOriginalMethod = class_getInstanceMethod(self, origSel);
            
            if (!inheritedOriginalMethod)
            {
                NSLog(@"Original method %@ not found for class %@", NSStringFromSelector(origSel), [self className]);
                
                return NO;
            }
        }
        
        if (!directAlternateMethod)
        {
            inheritedAlternateMethod = class_getInstanceMethod(self, altSel);
            
            if (!inheritedAlternateMethod)
            {
                NSLog(@"Alternate method %@ not found for class %@", NSStringFromSelector(altSel), [self className]);
                
                return NO;
            }
        }
        
        int hoisted_method_count = !directOriginalMethod && !directAlternateMethod ? 2 : 1;
        struct objc_method_list *hoisted_method_list = malloc(sizeof(struct objc_method_list) + (sizeof(struct objc_method) * (hoisted_method_count-1)));
        
        hoisted_method_list->method_count = hoisted_method_count;
        Method hoisted_method = hoisted_method_list->method_list;
        
        if (!directOriginalMethod)
        {
            bcopy(inheritedOriginalMethod, hoisted_method, sizeof(struct objc_method));
            directOriginalMethod = hoisted_method++;
        }
        
        if (!directAlternateMethod)
        {
            bcopy(inheritedAlternateMethod, hoisted_method, sizeof(struct objc_method));
            directAlternateMethod = hoisted_method;
        }
        
        class_addMethods(self, hoisted_method_list);
    }
    
    IMP temp = directOriginalMethod->method_imp;
    directOriginalMethod->method_imp = directAlternateMethod->method_imp;
    directAlternateMethod->method_imp = temp;
    
    return YES;
#endif
}

@end
