
/******************************************************************************
 *
 * � Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/

#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <sys/wait.h>
#include <boost/bind.hpp>
#include <string>
#include <map>
#include "pnserver.h"
#include "pnsession.h"

#define GUI_PORT        42331
#define OWN_PORT        42332
#define MULTIADDR_PORT  42333

using boost::asio::ip::tcp;
using boost::asio::ip::udp;

//---------------------------------------------------------------------------
pnserver::pnserver(boost::asio::io_service& io_service, unsigned short port)
        : io_service_(io_service),
        acceptor_(io_service)
{
    tcp::endpoint endpoint(tcp::v4(), port);
    acceptor_.open(endpoint.protocol());
    boost::asio::socket_base::reuse_address option(true);
    acceptor_.set_option(option);
    acceptor_.bind(endpoint);
    acceptor_.listen();
    syslog (LOG_NOTICE, "start");

    session_count       = 0;
    pnsession* session  = new pnsession(io_service_, this);
    master_session      = session;

    acceptor_.async_accept(session->socket(),
                           boost::bind(&pnserver::handle_accept, this, session,
                                       boost::asio::placeholders::error));
    // ping

    char buf[256];
    sprintf (buf, "%d", OWN_PORT);

    syslog (LOG_NOTICE, "ping");
    udp::socket bcast_socket(io_service_, udp::v4());
    udp::endpoint destination(boost::asio::ip::address::from_string("225.1.1.1"), MULTIADDR_PORT);
    bcast_socket.send_to(boost::asio::buffer(buf), destination);
    bcast_socket.close();



}

//---------------------------------------------------------------------------
pnserver::~pnserver()
{
    std::map <std::string, std::string>::iterator it;
    for (int i = 0; i < 3; i++)
    {
        for (it = app_lists_[i].begin(); it != app_lists_[i].end(); ++it)
        {
            app_lists_[i].erase (it);
        }
    }

    std::map <std::string, pnsession *>::iterator i;
    std::string name;
    for (i = sessions_.begin(); i != sessions_.end(); ++i)
    {
        name = i->first;
        delete i->second;
        sessions_.erase (i);
        syslog (LOG_NOTICE, "session %s closed", name.c_str());
    }

}

//---------------------------------------------------------------------------
void pnserver::inc_session_count(void)
{
    session_count ++;
    syslog (LOG_NOTICE, "Number session instances now %i.", session_count);}

void pnserver::dec_session_count(void)
{
    session_count --;
    syslog (LOG_NOTICE, "Number session instances now %i.", session_count);

    // If I ever delete th elast one, shut the server down
    if ( session_count == 0 )
    {
    syslog (LOG_NOTICE, "There are no other sessions.");
    throw std::logic_error ("stop request");
    }
}


//---------------------------------------------------------------------------
void pnserver::handle_accept(pnsession* session, const boost::system::error_code& error)
{
    if (!error)
    {
        syslog (LOG_NOTICE, "accept");

        session->start();
        session = new pnsession(io_service_, this);
        acceptor_.async_accept(session->socket(),
                               boost::bind(&pnserver::handle_accept, this, session,
                                           boost::asio::placeholders::error));
    }
    else
    {
        syslog (LOG_NOTICE, "accept error: %s", error.message().c_str());
        delete session;
    }
}

//---------------------------------------------------------------------------
std::string pnserver::get_list (const std::string& broker, int type)
{
    std::map <std::string, std::string>::iterator i = app_lists_[type].find (broker);
    if (i == app_lists_[type].end())
    {
        return std::string ("");
    }

    return i->second;
}

//---------------------------------------------------------------------------
void pnserver::set_list (const std::string& broker, const std::string& list, int type)
{
    syslog (LOG_NOTICE, "broker: %s   apps: %s", broker.c_str(), list.c_str());
    app_lists_[type][broker] = list;
}

//---------------------------------------------------------------------------
void pnserver::register_session (const std::string& session_name, pnsession * session)
{
    syslog (LOG_NOTICE, "register session: %s (%p)", session_name.c_str(), session);
    sessions_.insert(std::pair <std::string, pnsession *> (session_name, session));

    // Some debugging here
    syslog (LOG_NOTICE, "Session list has %i item(s) now.", sessions_.size());
}

//---------------------------------------------------------------------------
pnsession * pnserver::find_session (const std::string& session_name)
{
    syslog (LOG_NOTICE, "searching for session %s", session_name.c_str());
    std::map <std::string, pnsession *>::iterator i = sessions_.find (session_name);
    if (i == sessions_.end())
    {
        syslog (LOG_NOTICE, "not found");
        return NULL;
    }
    syslog (LOG_NOTICE, "found");
    return i->second;
}

//---------------------------------------------------------------------------
void pnserver::remove_session (pnsession * session)
{

    std::map <std::string, pnsession *>::iterator i;
    std::string name;

    syslog( LOG_NOTICE, "Attempting to remove session %p from list.", session );

    for (i = sessions_.begin(); i != sessions_.end(); ++i)
    {
        name = i->first;
        syslog (LOG_NOTICE, "Checking session: '%s' (%p)", name.c_str(), i->second);
        if (i->second == session)
        {
            name = i->first;
            sessions_.erase (i);
            syslog (LOG_NOTICE, "Found match, session %s closed", name.c_str());

            // Some debugging here
            syslog (LOG_NOTICE, "Session list has %i item(s) now.", sessions_.size());

                                    // If I deleted a session AND there are none left, then shut down
            if ((sessions_.size() == 0) && ( gui_sessions_.size() == 0 ))
                {
                syslog (LOG_NOTICE, "There are no other sessions.");
                throw std::logic_error ("stop request");
                }

            break;
        }
    }

}




//---------------------------------------------------------------------------
void sigchld_handler (int signo)
{
    pid_t pid;
    int stat;

    while ((pid = waitpid (-1, &stat, WNOHANG)) > 0);
    return;
}

//---------------------------------------------------------------------------
int main ()
{
    int i;
    struct sigaction sa;
    boost::asio::io_service io_service;

    openlog ("pnserver", LOG_PID, LOG_USER);

    // daemonize

    if ((i = fork()) < 0)
    {
        syslog (LOG_NOTICE, "error: %s", strerror (errno));
        exit (1);
    }
    else if (i)
    {
        exit (0);
    }
    setsid();
    close (0);
    close (1);
    close (2);
    i = open ("/dev/null", O_RDWR);
    dup (i);
    dup (i);
    chdir ("/");
    sigemptyset (&sa.sa_mask);
    sa.sa_handler = SIG_IGN;
    sa.sa_flags = 0;
    sigaction (SIGTSTP, &sa, NULL);
    sigaction (SIGTTIN, &sa, NULL);
    sigaction (SIGTTOU, &sa, NULL);
    sigaction (SIGHUP, &sa, NULL);
    sa.sa_handler = sigchld_handler;
    sa.sa_flags |= SA_RESTART;
    sigaction (SIGCHLD, &sa, NULL);

    // start server

    pnserver srvr (io_service, OWN_PORT);
    try
    {
        io_service.run();
    }
    catch (std::exception& e)
    {
        syslog (LOG_NOTICE, "an exception has been caught: %s", e.what());
    }

    syslog (LOG_NOTICE, "server stopped");
    closelog ();

    return 0;
}
