
/******************************************************************************
 *
 * � Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/

#include <stdio.h>
#include <jpeglib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "pnrdp.h"


int g_pnrdp_proto = 0;
int g_pnrdp_is_enabled = 0;

static uint32 gdip_total_size;
static uint32 gdip_total_emf_size;

static EMFP_OBJECT emfp_object_table[EMFP_OBJECT_TABLE_SIZE];

static unsigned char * pnrdp_order_buffer = NULL;
static size_t pnrdp_order_buffer_size = 0;  // current size of pnrdp_order_buffer

extern int pnrdp_decompress_image (unsigned char *cdata, size_t cdatasize, void **buffer);

extern void ui_pnrdp_set_region(Region r);

extern RD_BOOL g_gaenable;

// tries to read EMF+ PNRDP comment record from the stream
static int
pnrdp_check_comment(STREAM s)
{
    DEBUG_PNRDP(("Entering %s()...\n", __func__));
    EMFP_PNRDP_COMMENT comment;

    in_uint16_le(s, comment.type);
    if (comment.type != PNRDP_EMFP_COMMENT_TYPE)
        return -1;

    in_uint16_le(s, comment.flags);
    in_uint32_le(s, comment.size);
    if (comment.size < sizeof(EMFP_PNRDP_COMMENT))
        return -1;

    size_t datasize = comment.size - (sizeof(EMFP_PNRDP_COMMENT) - sizeof(char *));
    in_uint8p(s, comment.data, datasize);

    char *comment_str = (char *) xmalloc(datasize);
    comment_str[0] = '\0';
    int proto_version;
    if (sscanf((char *)comment.data, "%[a-zA-Z]%d", comment_str, &proto_version) <= 0)
        goto err_cond;
    if (strncmp(comment_str, PNRDP_EMFP_COMMENT_STRING, strlen(PNRDP_EMFP_COMMENT_STRING)) != 0)
        goto err_cond;
    if (proto_version != g_pnrdp_proto)
    {
//        goto err_cond;
	DEBUG_PNRDP(("server uses PNRDP%d proto", proto_version));
	g_pnrdp_proto = proto_version;
    }

    xfree(comment_str);
    return 0;

err_cond:

    xfree(comment_str);
    return -1;
}

static int
pnrdp_parse_emfp_object_record(STREAM s, EMFP_OBJECT_RECORD *record)
{
    DEBUG_PNRDP(("Entering %s()...\n", __func__));

    in_uint16_le(s, record->type);
    DEBUG_PNRDP((" record:  type: %#x ", record->type));
    if (record->type != EmfPlusObject)
    {
        error("PNRDP: cannot parse record - wrong type\n");
        return -1;
    }

    // get type and id
    in_uint16_le(s, record->object.flags);
    DEBUG_PNRDP(("flags: %#x ", record->object.flags));
    DEBUG_PNRDP(("object->c: %#x ", record->object.c));
    DEBUG_PNRDP(("object->type: %#x ", record->object.type));
    DEBUG_PNRDP(("object->id: %#x\n", record->object.id));

    switch (record->object.type)
    {
    case ObjectTypeRegion :
    case ObjectTypeImage :
        in_uint32_le(s, record->object.size);
//        in_uint32_le(s, record->object->datasize);
        if (record->object.size < sizeof(EMFP_OBJECT_RECORD))
            return -1;

        size_t datasize = record->object.size - (sizeof(EMFP_OBJECT_RECORD) - sizeof(char *));
        in_uint8p(s, record->object.data, datasize);
        break;
    default :
        error("PNRDP: this type of EMF+ object record is not supported\n");
        return -1;
    }

    DEBUG_PNRDP((" exit %s()\n", __func__));
    return 0;
}

static int
pnrdp_parse_set_clip_region_record(STREAM s, EMFP_SET_CLIP_REGION_RECORD *record)
{
    DEBUG_PNRDP(("Entering %s()...\n", __func__));
    in_uint16_le(s, record->type);
    if (record->type != EmfPlusSetClipRegion)
    {
        error("PNRDP: cannot parse a clip region record - wrong type\n");
        return -1;
    }
    in_uint16_le(s, record->flags);
    DEBUG_PNRDP((" flags: %#x, object_id: %#x, combine mode: %#x\n",
            record->flags, record->object_id, record->combine_mode));
    in_uint32_le(s, record->size);
//    in_uint32_le(s, record->datasize);
    DEBUG_PNRDP((" exit %s()\n", __func__));
    return 0;
}

static int
pnrdp_parse_draw_image_record(STREAM s, EMFP_DRAW_IMAGE_RECORD *record)
{
    DEBUG_PNRDP(("Entering %s()...\n", __func__));
    in_uint16_le(s, record->type);
    if (record->type != EmfPlusDrawImage)
    {
        error("PNRDP: cannot parse a draw image record - wrong type\n");
        return -1;
    }
    in_uint16_le(s, record->flags);
    DEBUG_PNRDP((" flags: %#x ", record->flags));
    DEBUG_PNRDP(("object_id: %#x ", record->object_id));
    DEBUG_PNRDP(("'c' bit: %#x\n", record->c));
    in_uint32_le(s, record->size);
//    in_uint32_le(s, record->datasize);
    in_uint32_le(s, record->image_attr_id);
    in_uint32_le(s, record->src_unit);
    in_uint32_le(s, record->src_rect.x);
    in_uint32_le(s, record->src_rect.y);
    in_uint32_le(s, record->src_rect.width);
    in_uint32_le(s, record->src_rect.height);
    if (record->c)
    {
        in_uint16_le(s, record->dst_rect.x);
        in_uint16_le(s, record->dst_rect.y);
        in_uint16_le(s, record->dst_rect.width);
        in_uint16_le(s, record->dst_rect.height);
    }
    else
    {
        in_uint32_le(s, record->dst_rectf.x);
        in_uint32_le(s, record->dst_rectf.y);
        in_uint32_le(s, record->dst_rectf.width);
        in_uint32_le(s, record->dst_rectf.height);
    }
    DEBUG_PNRDP((" exit %s()\n", __func__));
    return 0;
}

#define is_rectangle(a, b, c, d) \
( (a.y == b.y && b.x == c.x && c.y == d.y && d.x == a.x) || \
  (a.x == b.x && b.y == c.y && c.x == d.x && d.y == a.y) )

static void
pnrdp_create_and_apply_regions(EMFP_POINT *points,
                               EMFP_PATH_POINT_TYPE_NORMAL *types,
                               size_t pcount)
{
    int i = 0;

    Region r = XCreateRegion(); // an initial empty region

    while (i < pcount)
    {
        if ( is_rectangle (points[i], points[i+1], points[i+2], points[i+3]) &&
                types[i].type == PathPointTypeStart &&
                types[i+1].type == PathPointTypeLine &&
                types[i+2].type == PathPointTypeLine &&
                types[i+3].flags == PathPointTypeCloseSubpath)
        {
            XRectangle rect;
            rect.x = points[i].x;
            rect.y = points[i].y;
            rect.width = points[i+2].x - points[i].x;
            rect.height = points[i+2].y - points[i].y;
            i += 4;

            // add the rectangle to the main region
            XUnionRectWithRegion(&rect, r, r);
        }
        else
        {
            // not a rectangle, use XPolygonRegion
            int j = 0;
            // just bound the array with the maximum possible number of points
            XPoint *xpoints = xmalloc (sizeof(XPoint) * pcount);
            while (types[i].flags != PathPointTypeCloseSubpath)
            {
                xpoints[j].x = points[i].x;
                xpoints[j].y = points[i].y;
                ++i;
                ++j;
            }

            // create a region from the points, and add it to the main region
            Region r1 = XPolygonRegion(xpoints, j, EvenOddRule);
            XUnionRegion(r, r1, r);     // the third parameter is the result
            XDestroyRegion(r1);
            xfree (xpoints);
        }
    }

    ui_pnrdp_set_region(r);

    XDestroyRegion(r);
}

static void
pnrdp_set_clipping_region(EMFP_SET_CLIP_REGION_RECORD *record)
{
    int i;
    DEBUG_PNRDP(("Entering %s() - object_id: %#x\n", __func__, record->object_id));
    if (record->object_id > EMFP_OBJECT_TABLE_SIZE)
    {
        error ("PNRDP: cannot set clipping region - incorrect object ID\n");
        return;
    }

    EMFP_OBJECT *region_object = &emfp_object_table[record->object_id];

    if (record->combine_mode != CombineModeReplace)
    {
        warning ("PNRDP: warning: different combine mode - %d\n", record->combine_mode);
    }

    EMFP_REGION_OBJECT * region = (EMFP_REGION_OBJECT *) region_object->data;

    assert(region != NULL);

    // the region has region->ncount nodes,
    // which are either plain regions, or pathes, or child nodes
    EMFP_REGION_NODE *current_node = (EMFP_REGION_NODE *) &region->nodes;
    size_t current_node_size = 0, point_size = 0, point_type_size = 0;

    assert(current_node != NULL);

    for (i = 0; i < region->ncount + 1; ++i)
    {
        switch (current_node->type)
        {
        case RegionNodeDataTypePath :
        {
            // get the points
            EMFP_PATH *path = &current_node->path;

            assert(path != NULL);

            DEBUG_PNRDP(("Region: path, c %d  r %d  p %d,  pcount %d\n",
                    path->cbit, path->rbit, path->pbit, path->pcount));

            if (path->rbit)
            {
                point_type_size = sizeof(EMFP_PATH_POINT_TYPE_RLE);
            }
            else
            {
                point_type_size = sizeof(EMFP_PATH_POINT_TYPE_NORMAL);
            }

            if (path->pbit)
            {
                // the points are relative locations, and are specified
                // by EMFP_POINTR objects;
                point_size = sizeof (EMFP_POINTR);
		warning("Warning: unimplemented feature was requested: relative points!");

                // unimplemented...
            }
            else
            {
                // the points are absolute locations
                if (path->cbit)
                {
                    // the locations are specified by EMFP_POINT objects
                    point_size = sizeof (EMFP_POINT);

                    if (point_type_size == sizeof(EMFP_PATH_POINT_TYPE_NORMAL))
                    {
                        // here we should just ignore the compiler warning
                        // about dereferencing type-punned pointer (casting
                        // path->points to (EMFP_POINT *)), as with the current
                        // set of flags the path->points array is of exactly
                        // (EMFP_POINT *) type.
                        EMFP_POINT *points = (EMFP_POINT *)(&path->points);
                        EMFP_PATH_POINT_TYPE_NORMAL *types =
                            (EMFP_PATH_POINT_TYPE_NORMAL *)((char *)points + path->pcount * sizeof(EMFP_POINT));
                        pnrdp_create_and_apply_regions(points, types, path->pcount);
                    }
                }
                else
                {
                    // the locations are specified by EMFP_POINTF objects;
                    point_size = sizeof (EMFP_POINTF);
		warning("Warning: unimplemented feature was requested: floating points!");

                    // unimplemented...
                }
            }

            // update current_node_size
            current_node_size += (sizeof (EMFP_PATH)) -
                                 (sizeof (EMFP_PATH_POINT *) + sizeof (EMFP_PATH_POINT_TYPE *)) +
                                 path->pcount * (point_size + point_type_size);
        }
        current_node = (EMFP_REGION_NODE *)((char *)current_node + current_node_size);
        break;
        // for now I saw only Path regions in PNRDP messages...
        case RegionNodeDataTypeRect :
        default:
            error("PNRDP: regions of this type are not supported\n");
            // for now, just return...
            return;
        }
    }

    DEBUG_PNRDP((" exit %s()\n", __func__));
}

static void
pnrdp_draw_image(EMFP_DRAW_IMAGE_RECORD *record)
{
    DEBUG_PNRDP(("Entering %s() - object_id: %#x\n", __func__, record->object_id));
    EMFP_OBJECT *image_object = &emfp_object_table[record->object_id];
    EMFP_IMAGE_OBJECT *image = (EMFP_IMAGE_OBJECT *) image_object->data;

    if (image->type != ImageDataTypeBitmap)
    {
        error("PNRDP: this type of EMF+ image object is not supported\n");
        return;
    }
    EMFP_COMPRESSED_IMAGE *image_data = (EMFP_COMPRESSED_IMAGE *) &image->data;
    if (image_data->type != BitmapDataTypeCompressed)
    {
        error("PNRDP: only JPEG-compressed bitmaps are supported\n");
        return;
    }

    void *buffer = NULL;

    size_t cimagesize = image_object->size -
                        ( sizeof(EMFP_OBJECT) - sizeof(char *) ) -
                        ( sizeof(EMFP_IMAGE_OBJECT) - sizeof(char *) ) -
                        ( sizeof(EMFP_COMPRESSED_IMAGE) - sizeof(char *) );

    if (pnrdp_decompress_image((unsigned char *)&(image_data->cimage), cimagesize, &buffer) < 0)
    {
        error ("JPEG image decompression failed!\n");
        if (buffer != NULL)
            free (buffer);

        return;
    }

    int x, y, width, height;

    if (record->c)
    {
        // dest rect is compressed
        x = record->dst_rect.x;
        y = record->dst_rect.y;
        width = record->dst_rect.width;
        height = record->dst_rect.height;
    }
    else
    {
        x = record->dst_rectf.x;
        y = record->dst_rectf.y;
        width = record->dst_rectf.width;
        height = record->dst_rectf.height;
    }

    DEBUG_PNRDP(("Image: coordinates (xywh): %d %d %d %d\n", x, y, width, height));

    ui_pnrdp_paint_bitmap(x, y, width, height, width, height, buffer);

    free (buffer);

    DEBUG_PNRDP((" exit %s()\n", __func__));
}

static int
pnrdp_process_order(unsigned char * const pnrdp_order, size_t size)
{
    DEBUG_PNRDP(("Entering %s()...\n", __func__));

    struct stream S;
    S.data = pnrdp_order;
    S.size = size;
    S.end = S.p = S.data;

    STREAM s = &S;

    EMFP_OBJECT_RECORD record;

    // try to read PNRDP comment record
    if (pnrdp_check_comment(s) < 0)
    {
        error ("Wrong PNRDP order comment part\n");
        return -1;
    }

    // region object record
    if (pnrdp_parse_emfp_object_record(s, &record) < 0 ||
            record.object.type != ObjectTypeRegion)
    {
        return -1;
    }
    DEBUG_PNRDP(("parsed a record with object id %d\n", record.object.id));
    emfp_object_table[record.object.id] = record.object;

    // image (JPEG-compressed bitmap) object record
    if (pnrdp_parse_emfp_object_record(s, &record) < 0 ||
            record.object.type != ObjectTypeImage)
    {
        return -1;
    }
    DEBUG_PNRDP(("parsed a record with object id %d\n", record.object.id));
    emfp_object_table[record.object.id] = record.object;

    // parse SetClipRegion record
    EMFP_SET_CLIP_REGION_RECORD clip_record;
    if (pnrdp_parse_set_clip_region_record(s, &clip_record) < 0)
    {
        return -1;
    }

    // parse DrawImage record
    EMFP_DRAW_IMAGE_RECORD draw_record;
    if (pnrdp_parse_draw_image_record(s, &draw_record) < 0)
    {
        return -1;
    }

    // now apply the clipping region and draw the image
    pnrdp_set_clipping_region(&clip_record);
    pnrdp_draw_image(&draw_record);

    // reset clipping regions for rdesktop window to default
    ui_reset_clip();

    DEBUG_PNRDP((" exit %s()\n", __func__));
    return 0;
}


// This is the entry point; rdesktop code invokes us by calling this function

// Only PNRDP GDI+ orders are processed. They are processed
// if and only if PNRDP protocol is enabled.
int
pnrdp_process_altsec_order(STREAM s, uint8 order_flags)
{
    DEBUG_PNRDP(("\nEntering %s()...\n", __func__));
    int ret_code = 0;

    if (!g_gaenable)
    {
        error ("PNRDP alternate secondary orders support is disabled\n");
        ret_code = -1;
        goto done;
    }

    uint16 size;
    uint8 flags, type = order_flags >> 2;

    in_uint8(s, flags); // flags; not used
    in_uint16_le(s, size);

    switch (type)
    {
    case TS_ALTSEC_GDIP_FIRST:
    {
        in_uint32_le(s, gdip_total_size);
        in_uint32_le(s, gdip_total_emf_size);
        DEBUG_PNRDP(("TS_ALTSEC_GDIP_FIRST: %d %d %d\n", size, gdip_total_size, gdip_total_emf_size));

        pnrdp_order_buffer = realloc (pnrdp_order_buffer, gdip_total_size);
        in_uint8a(s, pnrdp_order_buffer, size);
        if (size != gdip_total_size)
        {
            pnrdp_order_buffer_size = size;
        }
        else
        {
            pnrdp_order_buffer_size = gdip_total_size;
        }
    }
    break;
    case TS_ALTSEC_GDIP_NEXT:
    {
        DEBUG_PNRDP(("TS_ALTSEC_GDIP_NEXT: %d\n", size));
        in_uint8a(s, pnrdp_order_buffer + pnrdp_order_buffer_size, size);
        pnrdp_order_buffer_size += size;
    }
    break;
    case TS_ALTSEC_GDIP_END:
    {
        DEBUG_PNRDP(("TS_ALTSEC_GDIP_END: %d\n", size));
        in_uint32_le(s, gdip_total_size);
        in_uint32_le(s, gdip_total_emf_size);
        in_uint8a(s, pnrdp_order_buffer + pnrdp_order_buffer_size, size);
        pnrdp_order_buffer_size = gdip_total_size;
    }
    break;
    default:
    {
        error ("This (%d) type of altsec drawing order is not supported\n", type);
        ret_code = -1;
        goto done;
    }
    }

    if (pnrdp_order_buffer_size == gdip_total_size)
    {
        // process the orders
        if (pnrdp_process_order(pnrdp_order_buffer, pnrdp_order_buffer_size) < 0)
        {
            ret_code = -1;
            goto done;
        }
    }

done:

    DEBUG_PNRDP((" exit %s(), ret_code %d\n\n", __func__, ret_code));
    return 0;
}

int
pnrdp_protocol_data_init()
{
    // actually the return code is absolutely useless here -
    //  xmalloc calls exit() if allocation is failed
    DEBUG_PNRDP(("\nEntering %s()...\n", __func__));
    pnrdp_order_buffer = xmalloc (1);
    DEBUG_PNRDP((" exit %s()\n", __func__));
    return 0;
}

void
pnrdp_protocol_data_deinit()
{
    DEBUG_PNRDP(("\nEntering %s()...\n", __func__));
    if (pnrdp_order_buffer != NULL)
    {
        free (pnrdp_order_buffer);
        pnrdp_order_buffer = NULL;
    }
    DEBUG_PNRDP((" exit %s()\n", __func__));
}
