
/******************************************************************************
 *
 * � Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/

/*  -*- c-basic-offset: 8 -*-
+   rdesktop: A Remote Desktop Protocol client.
+   Support for the USB channel
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+*/

#include "rdesktop.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdio.h>
#include <pthread.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#ifdef WITH_DEBUG_USBRDR
#define DEBUG_USBRDR(args) printf args;
#else
#define DEBUG_USBRDR(args)
#endif

#pragma pack (push, 1)

typedef struct _rdp_header {
	__int64_t id_connection;
	__int32_t packet_size;
} RDP_HEADER, *PRDP_HEADER;

#pragma pack (pop)

static void usbrdr_send(const void *output, size_t len);

static VCHANNEL *usbrdr_channel;
static pthread_mutex_t channel_send_lock;

#define USBRDR_TCPPORT 5206
#define USBRDR_RECV_BUFFER_SIZE 100000

typedef struct _CONNECTION_CONTEXT
{
	struct _CONNECTION_CONTEXT *next;
	int socket;
	pthread_t *thread;
	int stop;
	int disconnecting;
} CONNECTION_CONTEXT;

static CONNECTION_CONTEXT *connection_list = NULL;
static pthread_mutex_t connection_list_lock;

static RDP_HEADER usbrdr_header;
static CONNECTION_CONTEXT *usbrdr_context = NULL;


/* find connection context by connection id */
static CONNECTION_CONTEXT *
usbrdr_connection_find(__int64_t id_connection)
{
	CONNECTION_CONTEXT *ptr;

	pthread_mutex_lock(&connection_list_lock);

	ptr = connection_list;

	while (ptr)
	{
		if (ptr->socket == id_connection) break;
		ptr = ptr->next;
	}

	pthread_mutex_unlock(&connection_list_lock);

	return ptr;
}

/* remove connection context by connection id */
static void 
usbrdr_connection_remove(CONNECTION_CONTEXT *ctx)
{
	CONNECTION_CONTEXT *ptr, *prevptr;

	pthread_mutex_lock(&connection_list_lock);

	prevptr = NULL;
	ptr = connection_list;

	while (ptr)
	{
		if (ptr == ctx) break;
		prevptr = ptr;
		ptr = ptr->next;
	}

	if (ptr)
	{
		if(prevptr) prevptr->next = ptr->next; else connection_list = ptr->next;
	}

	pthread_mutex_unlock(&connection_list_lock);
}

/* add connection context to the list */
static void 
usbrdr_connection_add(CONNECTION_CONTEXT *ptr)
{
	pthread_mutex_lock(&connection_list_lock);
	ptr->next = connection_list;
	connection_list = ptr;
	pthread_mutex_unlock(&connection_list_lock);
}

static void *
socket_handler_function(void *data)
{
	CONNECTION_CONTEXT *context = (CONNECTION_CONTEXT*)data;
	void *buf;

	buf = xmalloc(USBRDR_RECV_BUFFER_SIZE);
	if(buf == NULL)
	{
		error("USBRDR: THREAD %08X no memory for recv buffer\n", context->socket);
		return NULL;
	}

	while (!context->stop)
	{
		int nrecv;
		RDP_HEADER header;

		nrecv = recv(context->socket, buf, USBRDR_RECV_BUFFER_SIZE,0);

		if (nrecv <= 0) 
		{
			DEBUG_USBRDR(("USBRDR: THREAD %08X recv error\n", context->socket));
			break;
		}

		header.id_connection = context->socket;
		header.packet_size = nrecv;

		pthread_mutex_lock(&channel_send_lock);
		usbrdr_send(&header, sizeof(header));
		usbrdr_send(buf, nrecv);
		pthread_mutex_unlock(&channel_send_lock);
	}

	xfree(buf);

	if (!context->disconnecting)
	{
		RDP_HEADER header;

		DEBUG_USBRDR(("USBRDR: THREAD %08X sending disconnect for\n", context->socket));

		header.id_connection = context->socket;
		header.packet_size = -1;

		pthread_mutex_lock(&channel_send_lock);
		usbrdr_send(&header, sizeof(header));
		pthread_mutex_unlock(&channel_send_lock);
	}

	shutdown(context->socket, SHUT_WR);
	close(context->socket);

	DEBUG_USBRDR(("USBRDR: THREAD %08X terminating\n", context->socket));
	return NULL;
}


static void 
usbrdr_process_cleanup()
{
	/* close all connections */

	while(connection_list)
	{
		CONNECTION_CONTEXT *context = connection_list;

		connection_list = connection_list->next;

		DEBUG_USBRDR(("USBRDR: cleaning up connection %08X...\n", context->socket));

		context->disconnecting = 1;
		context->stop = 1;
		shutdown(context->socket, SHUT_RD);
		pthread_join(*context->thread, NULL);

		DEBUG_USBRDR(("USBRDR: connection %08X cleaned up\n", context->socket));

		xfree(context->thread);
		xfree(context);	
	}
}

static void 
usbrdr_process_connect()
{
	int				s = -1;
	pthread_t 			*socket_handler	= NULL;
	CONNECTION_CONTEXT	*context 		= NULL;


	do /*while(0)*/
	{
		RDP_HEADER reply_header;
	    struct sockaddr_in sin;
		int opt = 1;

		reply_header.id_connection = 0;
		reply_header.packet_size = -1;

		context = xmalloc(sizeof(*context));

		if (context == NULL)
		{
			error("USBRDR: no memory for connection context\n");
			break;
		}

		socket_handler = xmalloc(sizeof(*socket_handler));

		if (socket_handler == NULL)
		{
			error("USBRDR: no memory for thread\n");
			break;
		}

		s = socket(PF_INET, SOCK_STREAM, 0);

		if (s < 0) 
		{
			error("USBRDR: cannot create socket\n");
			break;
		}
		
		if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt)) < 0)
		{
			warning("USBRDR: cannot enable TCP_NODELAY option\n");
		}

		sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
		sin.sin_port = htons(USBRDR_TCPPORT);
		sin.sin_family = AF_INET;

		if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) != 0) 
		{
			error("USBRDR: cannot connect to local server");
			break;
		}

		context->socket = s;
		context->thread = socket_handler;
		context->stop = 0;
		context->disconnecting = 0;

		pthread_mutex_lock(&channel_send_lock);
	
		if (pthread_create(socket_handler, NULL, (void *(*)(void *))socket_handler_function, context))
		{
			pthread_mutex_unlock(&channel_send_lock);
			error("USBRDR: cannot create thread\n");
			break;
		}

		reply_header.id_connection = s;
		reply_header.packet_size = 0;

		usbrdr_send(&reply_header, sizeof(reply_header));

		usbrdr_connection_add(context);

		pthread_mutex_unlock(&channel_send_lock);


		s = -1;
		socket_handler = NULL;
		context = NULL;


		DEBUG_USBRDR(("USBRDR: new connection successful, connectionid is %08X\n", (__int32_t)reply_header.id_connection));
	} while(0);

	/* cleanup on error */
	if (s > 0) 
	{
		shutdown(s, SHUT_RDWR);
		close(s);
	}
	if (socket_handler) xfree(socket_handler);
	if (context) xfree(context);
}

static void 
usbrdr_process_disconnect()
{
	CONNECTION_CONTEXT *context = usbrdr_connection_find(usbrdr_header.id_connection);	

	if (!context) return;

	DEBUG_USBRDR(("USBRDR: disconnecting %08X...\n", (__int32_t)usbrdr_header.id_connection));

	context->disconnecting = 1;
	context->stop = 1;
	shutdown(context->socket, SHUT_RD);
	pthread_join(*context->thread, NULL);

	usbrdr_connection_remove(context);

	xfree(context->thread);
	xfree(context);
	DEBUG_USBRDR(("USBRDR: connection %08X is now disconnected\n", (__int32_t)usbrdr_header.id_connection));
}

static int 
usbrdr_process_data_header()
{
	usbrdr_context = usbrdr_connection_find(usbrdr_header.id_connection);	
	return 1;
}

static int 
usbrdr_process_data_chunk(void *data, size_t length)
{
	size_t tosend = (usbrdr_header.packet_size > length) ? length : usbrdr_header.packet_size;

	usbrdr_header.packet_size -= tosend;

	if (usbrdr_context)
	{
		if( send(usbrdr_context->socket, data, tosend, 0) <= 0 )
		{
			usbrdr_context->stop = 1;
			shutdown(usbrdr_context->socket, SHUT_RD);
			pthread_join(*usbrdr_context->thread, NULL);
			usbrdr_connection_remove(usbrdr_context);
			xfree(usbrdr_context->thread);
			xfree(usbrdr_context);
		}
	}
	return (usbrdr_header.packet_size > 0);
}

static void 
usbrdr_process_chunk(void *data, size_t length)
{
	static int rdpstate = 0;

	switch(rdpstate)
	{
		case 0: // waiting for rdp_header

			if( length == sizeof(usbrdr_header) )
			{
				memcpy(&usbrdr_header, data, sizeof(usbrdr_header));

				if( usbrdr_header.id_connection == 0 && usbrdr_header.packet_size == 0 )
				{
					DEBUG_USBRDR(("USBRDR: connect request\n"));
					usbrdr_process_connect();

				} else if( usbrdr_header.id_connection != 0 && usbrdr_header.packet_size == -1 )
				{
					DEBUG_USBRDR(("USBRDR: disconnect request for %08X\n", (__int32_t)usbrdr_header.id_connection));
					usbrdr_process_disconnect();

				} else if( usbrdr_header.id_connection == 0 && usbrdr_header.packet_size == -1 )
				{
					DEBUG_USBRDR(("USBRDR: cleanup request\n"));
					usbrdr_process_cleanup();

				} else
				{
					DEBUG_USBRDR(("USBRDR: data header\n"));
					if (usbrdr_process_data_header()) rdpstate = 1; // wait for data
				} 
			} else
			{
				error("USBRDR: error, wrong header size\n");
			}
			break;

		case 1: // waiting for data
			DEBUG_USBRDR(("USBRDR: data chunk %d bytes\n", usbrdr_header.packet_size));
			if( !usbrdr_process_data_chunk(data, length) ) rdpstate = 0; // wait for header
			break;

		default: // something wrong
			break;			
	}
}

/* Process new data from the virtual channel */
static void
usbrdr_process(STREAM s)
{
	unsigned int len;

	len = s->end - s->p;

#if 0
	printf("USBRDR: recv %d\n", len);
	hexdump(s->p, len);
#endif

	usbrdr_process_chunk(s->p, len);
}

/* Initialize this module: Register the usbrdr channel */
RD_BOOL
usbrdr_init(void)
{
	if (pthread_mutex_init(&channel_send_lock, NULL) != 0)
	{
		error("USBRDR: can not initialize channel access mutex\n");
		return 0;
	}

	if (pthread_mutex_init(&connection_list_lock, NULL) != 0)
	{
		error("USBRDR: can not initialize connection list access mutex\n");
		return 0;
	}

	usbrdr_channel =
		channel_register("QSTRDR", CHANNEL_OPTION_INITIALIZED | CHANNEL_OPTION_ENCRYPT_RDP,
				 usbrdr_process);

	DEBUG_USBRDR(("USBRDR: init complete\n"));

	return (usbrdr_channel != NULL);
}

/* Send data to channel */
static void 
usbrdr_send(const void *data, size_t len)
{
	STREAM s;

	s = channel_init(usbrdr_channel, len);

	out_uint8p(s, data, len) s_mark_end(s);

#if 0
	printf("USBRDR: send %d\n", len);
	hexdump(s->channel_hdr+8, s->end - s->channel_hdr-8);
#endif

	channel_send(s, usbrdr_channel);
}
